"""The model decorator."""

import typing
import numpy as np
import xarray as xr

from glotaran.analysis.result import Result
from glotaran.parameter import ParameterGroup
from glotaran.parse.register import register_model

from .dataset_descriptor import DatasetDescriptor
from .model import Model
from .util import wrap_func_as_method


MatrixFunction = typing.Callable[
    [typing.Type[DatasetDescriptor], xr.Dataset],
    typing.Tuple[typing.List[str], np.ndarray]]
"""A `MatrixFunction` calculates the matrix for a model."""

IndexDependedMatrixFunction = typing.Callable[
    [typing.Type[DatasetDescriptor], xr.Dataset, typing.Any],
    typing.Tuple[typing.List[str], np.ndarray]]
"""A `MatrixFunction` calculates the matrix for a model."""

GlobalMatrixFunction = typing.Callable[
    [typing.Type[DatasetDescriptor], np.ndarray],
    typing.Tuple[typing.List[str], np.ndarray]]
"""A `GlobalMatrixFunction` calculates the global matrix for a model."""

ConstrainMatrixFunction = typing.Callable[
    [typing.Type[Model], ParameterGroup, typing.List[str], np.ndarray, float],
    typing.Tuple[typing.List[str], np.ndarray]]
"""A `ConstrainMatrixFunction` applies constraints on a matrix."""

FinalizeFunction = typing.Callable[
    [typing.Type[Model], Result], None]
"""A `FinalizeFunction` gets called after optimization."""

PenaltyFunction = typing.Callable[
    [typing.Type[Model], ParameterGroup, typing.List[str], np.ndarray, any],
    np.ndarray]
"""A `PenaltyFunction` calculates additional penalties for the optimization."""


def model(model_type: str,
          attributes: typing.Dict[str, typing.Any] = {},
          dataset_type: typing.Type[DatasetDescriptor] = DatasetDescriptor,
          megacomplex_type: typing.Any = None,
          matrix: typing.Union[MatrixFunction, IndexDependedMatrixFunction] = None,
          global_matrix: GlobalMatrixFunction = None,
          matrix_dimension: str = None,
          global_dimension: str = None,
          has_matrix_constraints_function: typing.Callable[[typing.Type[Model]], bool] = None,
          constrain_matrix_function: ConstrainMatrixFunction = None,
          has_additional_penalty_function: typing.Callable[[typing.Type[Model]], bool] = None,
          additional_penalty_function: PenaltyFunction = None,
          finalize_data_function: FinalizeFunction = None,
          grouped: typing.Union[bool, typing.Callable[[typing.Type[Model]], bool]] = False,
          index_dependend: typing.Union[bool, typing.Callable[[typing.Type[Model]], bool]] = False,
          ) -> typing.Callable:
    """The `@model` decorator is intended to be used on subclasses of :class:`glotaran.model.Model`.
    It creates properties for the given attributes as well as functions to add access them. Also it
    adds the functions (e.g. for `matrix`) to the model ansures they are added wrapped in a correct
    way.

    Parameters
    ----------
    model_type :
        Human readable string used by the parser to identify the correct model.
    attributes :
        A dictionary of attribute names and types. All types must be decorated with the
        :func:`glotaran.model.model_attribute` decorator.
    dataset_type :
     A subclass of :class:`DatasetDescriptor`
    megacomplex_type :
        A class for the model megacomplexes. The class must be decorated with the
        :func:`glotaran.model.model_attribute` decorator.
    matrix :
        A function to calculate the matrix for the model.
    global_matrix :
        A function to calculate the global matrix for the model.
    matrix_dimension :
        The name of model matrix row dimension.
    global_dimension :
        The name of model global matrix row dimension.
    constrain_matrix_function :
        A function to constrain the global matrix for the model.
    None
    additional_penalty_function : PenaltyFunction
        A function to calculate additional penalties when optimizing the model.
    finalize_data_function :
        A function to finalize data after optimization.
    allow_grouping :
        If `True`, datasets can can be grouped along the global dimension.
    """

    def decorator(cls):

        setattr(cls, '_model_type', model_type)
        setattr(cls, 'finalize_data', finalize_data_function)

        if has_matrix_constraints_function:
            if not constrain_matrix_function:
                raise Exception('Model implements `has_matrix_constraints_function` '
                                'but not `constrain_matrix_function`')
            has_c_mat = wrap_func_as_method(
                cls, name='has_matrix_constraints_function'
            )(has_matrix_constraints_function)
            c_mat = wrap_func_as_method(
                cls, name='constrain_matrix_function'
            )(constrain_matrix_function)
            setattr(cls, 'has_matrix_constraints_function', has_c_mat)
            setattr(cls, 'constrain_matrix_function', c_mat)
        else:
            setattr(cls, 'has_matrix_constraints_function', None)
            setattr(cls, 'constrain_matrix_function', None)

        if has_additional_penalty_function:
            if not additional_penalty_function:
                raise Exception('Model implements `has_additional_penalty_function`'
                                'but not `additional_penalty_function`')
            has_pen = wrap_func_as_method(
                cls, name='has_additional_penalty_function'
            )(has_additional_penalty_function)
            pen = wrap_func_as_method(
                cls, name='additional_penalty_function'
            )(additional_penalty_function)
            setattr(cls, 'additional_penalty_function', pen)
            setattr(cls, 'has_additional_penalty_function', has_pen)
        else:
            setattr(cls, 'has_additional_penalty_function', None)
            setattr(cls, 'additional_penalty_function', None)

        if not callable(grouped):
            def group_fun(model):
                return grouped
        else:
            group_fun = grouped
        setattr(cls, 'grouped', group_fun)

        if not callable(index_dependend):
            def index_dep_fun(model):
                return index_dependend
        else:
            index_dep_fun = index_dependend
        setattr(cls, 'index_dependend', index_dep_fun)

        mat = wrap_func_as_method(cls, name='matrix')(matrix)
        mat = staticmethod(mat)
        setattr(cls, 'matrix', mat)
        setattr(cls, 'matrix_dimension', matrix_dimension)

        if global_matrix:
            g_mat = wrap_func_as_method(cls, name='global_matrix')(global_matrix)
            g_mat = staticmethod(g_mat)
            setattr(cls, 'global_matrix', g_mat)
        else:
            setattr(cls, 'global_matrix', None)
        setattr(cls, 'global_dimension', global_dimension)

        if not hasattr(cls, '_glotaran_model_attributes'):
            setattr(cls, '_glotaran_model_attributes', {})
        else:
            setattr(cls, '_glotaran_model_attributes',
                    getattr(cls, '_glotaran_model_attributes').copy())

        attributes['megacomplex'] = megacomplex_type
        attributes['dataset'] = dataset_type

        # Set annotations and methods for attributes
        for attr_name, attr_type in attributes.items():
            getattr(cls, '_glotaran_model_attributes')[attr_name] = None
            attr_prop = _create_property_for_attribute(cls, attr_name, attr_type)
            setattr(cls, attr_name, attr_prop)

            if getattr(attr_type, '_glotaran_has_label'):
                get_item = _create_get_func(cls, attr_name, attr_type)
                setattr(cls, get_item.__name__, get_item)
                set_item = _create_set_func(cls, attr_name, attr_type)
                setattr(cls, set_item.__name__, set_item)

            else:
                add_item = _create_add_func(cls, attr_name, attr_type)
                setattr(cls, add_item.__name__, add_item)

        init = _create_init_func(cls, attributes)
        setattr(cls, '__init__', init)

        register_model(model_type, cls)

        return cls

    return decorator


def _create_init_func(cls, attributes):

    @wrap_func_as_method(cls)
    def __init__(self):
        for attr_name, attr_item in attributes.items():
            if getattr(attr_item, '_glotaran_has_label'):
                setattr(self, f'_{attr_name}', {})
            else:
                setattr(self, f'_{attr_name}', [])
        super(cls, self).__init__()

    return __init__


def _create_add_func(cls, name, type):

    @wrap_func_as_method(cls, name=f'add_{name}')
    def add_item(self, item: type):
        f'''Adds an `{type.__name__}` object.

        Parameters
        ----------
        item :
            The `{type.__name__}` item.
        '''

        if not isinstance(item, type):
            if not hasattr(type, "_glotaran_model_attribute_typed") or \
               not isinstance(item, tuple(type._glotaran_model_attribute_types.values())):
                raise TypeError
        getattr(self, f'_{name}').append(item)
    return add_item


def _create_get_func(cls, name, type):

    @wrap_func_as_method(cls, name=f'get_{name}')
    def get_item(self, label: str) -> type:
        f'''
        Returns the `{type.__name__}` object with the given label.

        Parameters
        ----------
        label :
            The label of the `{type.__name__}` object.
        '''
        return getattr(self, f'_{name}')[label]

    return get_item


def _create_set_func(cls, name, type):

    @wrap_func_as_method(cls, name=f'set_{name}')
    def set_item(self, label: str, item: type):
        f'''
        Sets the `{type.__name__}` object with the given label with to the item.

        Parameters
        ----------
        label :
            The label of the `{type.__name__}` object.
        item :
            The `{type.__name__}` item.
        '''

        if not isinstance(item, type):
            if not hasattr(type, "_glotaran_model_attribute_typed") or \
               not isinstance(item, tuple(type._glotaran_model_attribute_types.values())):
                raise TypeError
        getattr(self, f'_{name}')[label] = item

    return set_item


def _create_property_for_attribute(cls, name, type):

    return_type = typing.Dict[str, type] if hasattr(type, '_glotaran_has_label') \
            else typing.List[type]

    doc_type = 'dictionary' if hasattr(type, '_glotaran_has_label') else 'list'

    @property
    @wrap_func_as_method(cls, name=f'{name}')
    def attribute(self) -> return_type:
        f'''A {doc_type} containing {type.__name__}'''
        return getattr(self, f'_{name}')

    return attribute



import click
from click import echo, pause, prompt

from . import util

@click.option('--name', '-n', default=None, type=str, show_default=True,
              help='Name of the datagroup to export.')
@click.option('--select', '-s', default=None, type=(str, util.VALORRANGEORLIST), show_default=True,
              multiple=True, help="Selection of data. Example --select DIMENSION_NAME VALUE "
              "where VALUE can be a scalar, a list like '[V1,V2,V3]' or a tuple like '(MIN,MAX)'"
              )
@click.option('--out', '-o', default=None, type=click.Path(dir_okay=False), show_default=True,
              help='Path to the output file.')
@click.argument("filename", type=click.Path(exists=True))
def export(filename: str, select, out: str, name: str):
    """Exports data from netCDF4 to ascii."""

    echo(f"Opening dataset at {filename}")

    dataset = util.load_dataset(filename, dtype='nc')

    if name is None:
        name = util.select_name(filename, dataset)

    stop = False
    data = dataset[name]

    for sel in select:
        data = util.select_data(data, sel[0], sel[1])

    if out is not None:
        if len(data.shape) > 2:
            pause('Cannot export data with more than 2 dimensions to ASCII')
        else:
            util.write_data(data, out)
            stop = True

    while not stop:

        echo(f"Selected dataset '{name}'.")
        echo(f"\nDataset Content\n\n{data}\n")

        choice = prompt('How to proceed', default='exit',
                        type=click.Choice(["exit", "save", "select", "change", "help"]))
        if choice == 'exit':
            stop = True
        elif choice == 'change':
            name = util.select_name(filename, dataset)
            data = dataset[name]
        elif choice == 'select':
            cont = True
            while cont:
                dims = [d for d in data.dims]
                choice = prompt('Please select a dimension or action', default='back',
                                type=click.Choice(dims + ['reset', 'back']))
                if choice == 'back':
                    cont = False
                elif choice == 'reset':
                    data = dataset[name]
                else:
                    dim = choice
                    choice = prompt(
                        "Please select a value. Type 2 values sperated by ',' to select a range.",
                        default='back', type=util.VALORRANGEORLIST
                    )
                    if not choice == 'back':
                        try:
                            data = util.select_data(data, dim, choice)
                        except ValueError as e:
                            echo(e, err=True)
                            continue

        elif choice == 'save':
            if len(data.shape) > 2:
                pause('Cannot export data with more than 2 dimensions to ASCII')
            else:
                cont = True
                change = False
                while cont:
                    if out is None or change:
                        out = prompt('Please type in a name for the output file.',
                                     default='out.csv', type=click.Path(dir_okay=False))
                        change = False
                    choice = prompt(f"The filepath of the ouput file is '{out}'. Save the data?",
                                    default='yes', type=click.Choice(['yes', 'no', 'change']))
                    if choice == 'change':
                        change = True
                    else:
                        cont = False
                        if choice == 'yes':
                            util.write_data(data, out)
    echo('Good-bye, have a nice day!')



"""The parameter group class"""

from lmfit import Parameters
from math import log
import copy
import csv
import numpy as np
import pandas as pd
import pathlib
import typing
import yaml

from .parameter import Parameter


class ParameterNotFoundException(Exception):
    """Raised when a Parameter is not found in the Group."""
    def __init__(self, path, label):
        super(ParameterNotFoundException, self).__init__(
            f"Cannot find parameter {'.'.join(path)}.{label}")


class ParameterGroup(dict):

    def __init__(self, label: str = None):
        """Represents are group of parameters. Can contain other groups, creating a
        tree-like hirachy.

        Parameters
        ----------
        label :
            The label of the group.
        """

        self._label = label
        self._parameters = {}
        self._root = None
        super(ParameterGroup, self).__init__()

    @classmethod
    def from_parameter_dict(cls, parameter: Parameters):
        """Creates a :class:`ParameterGroup` from an lmfit.Parameters dictionary

        Parameters
        ----------
        parameter :
            A lmfit.Parameters dictionary
        """

        root = cls(None)
        for lbl, param in parameter.items():
            lbl = lbl.split("_")
            if len(lbl) == 2:
                # it is a root param
                param = Parameter.from_parameter(lbl.pop(), param)
                root.add_parameter(param)
                continue

            # remove root
            lbl.pop(0)

            top = root
            while len(lbl) != 0:
                group = lbl.pop(0)
                if group in top:
                    if len(lbl) == 1:
                        param = Parameter.from_parameter(lbl.pop(), param)
                        top[group].add_parameter(param)
                    else:
                        top = top[group]
                else:
                    group = ParameterGroup(group)
                    top.add_group(group)
                    if len(lbl) == 1:
                        param = Parameter.from_parameter(lbl.pop(), param)
                        group.add_parameter(param)
                    else:
                        top = group
        return root

    @classmethod
    def from_dict(cls,
                  parameter: typing.Dict[str, typing.Union[typing.Dict, typing.List]],
                  label="p") -> "ParameterGroup":
        """Creates a :class:`ParameterGroup` from a dictionary.

        Parameters
        ----------
        parameter :
            The parameter dictionary.
        label :
            The label of root group.
        """
        root = cls(label)
        for label, item in parameter.items():
            label = str(label)
            if isinstance(item, dict):
                root.add_group(cls.from_dict(item, label=label))
            if isinstance(item, list):
                root.add_group(cls.from_list(item, label=label))
        return root

    @classmethod
    def from_list(cls,
                  parameter: typing.List[typing.Union[float, typing.List]],
                  label="p") -> "ParameterGroup":
        """Creates a :class:`ParameterGroup` from a list.

        Parameters
        ----------
        parameter :
            The parameter list.
        label :
            The label of the root group.
        """
        root = cls(label)

        # get defaults
        defaults = None
        for item in parameter:
            if isinstance(item, dict):
                defaults = item
                break

        for item in parameter:
            if isinstance(item, str):
                try:
                    item = float(item)
                except Exception:
                    pass
            if isinstance(item, (float, int, list)):
                root.add_parameter(Parameter.from_list_or_value(item, default_options=defaults))
        return root

    @classmethod
    def known_formats(cls) -> typing.Dict[str, typing.Callable]:
        return {
            'csv': cls.from_csv,
            'yml': cls.from_yaml_file,
            'yaml': cls.from_yaml_file,
        }

    @classmethod
    def from_file(cls, filepath: str, fmt: str = None):
        if fmt is None:
            path = pathlib.Path(filepath)
            fmt = path.suffix[1:] if path.suffix != '' else 'yml'
        if fmt not in cls.known_formats():
            raise Exception(f"Unknown parameter format '{format}'. "
                            f"Valid Formats are {cls.known_formats().keys()}.")
        return cls.known_formats()[fmt](filepath)

    @classmethod
    def from_yaml_file(cls, filepath: str) -> "ParameterGroup":
        """Creates a :class:`ParameterGroup` from a YAML file.

        Parameters
        ----------
        filepath :
            The path to the YAML file.
        """

        with open(filepath) as f:
            cls = cls.from_yaml(f)
        return cls

    @classmethod
    def from_yaml(cls, yaml_string: str):
        """Creates a :class:`ParameterGroup` from a YAML string.

        Parameters
        ----------
        yaml_string :
            The YAML string with the parameters.
        """
        items = yaml.load(yaml_string, Loader=yaml.FullLoader)
        if isinstance(items, list):
            cls = cls.from_list(items)
        else:
            cls = cls.from_dict(items)
        return cls

    @classmethod
    def from_csv(cls, filepath: str, delimiter: str = '\t'):
        """Creates a :class:`ParameterGroup` from a CSV file.

        Parameters
        ----------
        filepath :
            The path to the CSV file.
        delimiter : str
            The delimiter of the CSV file.
        """

        root = cls()
        df = pd.read_csv(filepath, sep=delimiter)

        for i, label in enumerate(df['label']):
            label = label.split('.')
            if len(label) == 1:
                p = Parameter(label=label.pop())
                p.value = df['value'][i]
                p.stderr = df['stderr'][i]
                p.min = df['min'][i]
                p.max = df['max'][i]
                p.vary = df['vary'][i]
                p.non_neg = df['non-negative'][i]
                root.add_parameter(p)
                continue

            top = root
            while len(label) != 0:
                group = label.pop(0)
                if group in top:
                    if len(label) == 1:
                        p = Parameter(label=label.pop())
                        p.value = df['value'][i]
                        p.stderr = df['stderr'][i]
                        p.min = df['min'][i]
                        p.max = df['max'][i]
                        p.vary = df['vary'][i]
                        p.non_neg = df['non-negative'][i]
                        top[group].add_parameter(p)
                    else:
                        top = top[group]
                else:
                    group = ParameterGroup(group)
                    top.add_group(group)
                    if len(label) == 1:
                        p = Parameter(label=label.pop())
                        p.value = df['value'][i]
                        p.stderr = df['stderr'][i]
                        p.min = df['min'][i]
                        p.max = df['max'][i]
                        p.vary = df['vary'][i]
                        p.non_neg = df['non-negative'][i]
                        group.add_parameter(p)
                    else:
                        top = group
        return root

    def to_csv(self, filename: str, delimiter: str = '\t'):
        """Writes a :class:`ParameterGroup` to a CSV file.

        Parameters
        ----------
        filepath :
            The path to the CSV file.
        delimiter : str
            The delimiter of the CSV file.
        """

        with open(filename, mode='w') as parameter_file:
            parameter_writer = csv.writer(parameter_file, delimiter=delimiter)
            parameter_writer.writerow(
                ['label', 'value', 'min', 'max', 'vary', 'non-negative', 'stderr']
            )

            for (label, p) in self.all():
                parameter_writer.writerow(
                    [label, p.value, p.min, p.max, p.vary, p.non_neg, p.stderr]
                )

    def add_parameter(self, parameter: Parameter):
        """Adds a :class:`Parameter` to the group.

        Parameters
        ----------
        parameter :
            The parameter to add.
        """
        if not isinstance(parameter, list):
            parameter = [parameter]
        if any(not isinstance(p, Parameter) for p in parameter):
            raise TypeError("Parameter must be  instance of"
                            " glotaran.model.Parameter")
        for p in parameter:
            p.index = len(self._parameters) + 1
            if p.label is None:
                p.label = f"{p.index}"
            p.full_label = f'{self.label}.{p.label}' if self.label else p.label
            self._parameters[p.label] = p

    def add_group(self, group: 'ParameterGroup'):
        """Adds a :class:`ParameterGroup` to the group.

        Parameters
        ----------
        group :
            The group to add.
        """
        if not isinstance(group, ParameterGroup):
            raise TypeError("Group must be glotaran.model.ParameterGroup")
        group.set_root(self)
        self[group.label] = group

    def set_root(self, root: 'ParameterGroup'):
        """Sets the root of the group.

        Parameters
        ----------
        root :
            The new root of the group.
        """
        self._root = root

    def get_nr_roots(self) -> int:
        """Returns the number of roots of the group."""
        n = 0
        root = self._root
        while root is not None:
            n += 1
            root = root._root
        return n

    @property
    def label(self) -> str:
        """Label of the group """
        return self._label

    def groups(self) -> typing.Generator['ParameterGroup', None, None]:
        """Returns a generator over all groups and their subgroups."""
        for group in self:
            for l in group.groups():
                yield l

    def has(self, label: str) -> bool:
        """Checks if a parameter with the given label is in the group or in a subgroup.

        Parameters
        ----------
        label :
            The label of the parameter.
        """

        try:
            self.get(label)
            return True
        except Exception:
            return False

    def get(self, label: str) -> Parameter:
        """Gets a :class:`Parameter` by it label.

        Parameters
        ----------
        label :
            The label of the parameter.
        """

        # sometimes the spec parser delivers the labels as int
        label = str(label)

        path = label.split(".")
        label = path.pop()

        group = self
        for l in path:
            try:
                group = group[l]
            except KeyError:
                raise ParameterNotFoundException(path, label)
        try:
            return group._parameters[label]
        except KeyError:
            raise ParameterNotFoundException(path, label)

    def all(self, root: str = None, seperator: str = ".") \
            -> typing.Generator[typing.Tuple[str, Parameter], None, None]:
        """Returns a generator over all parameter in the group and it's subgroups together with
        their labels.

        Parameters
        ----------
        root :
            The label of the root group
        seperator:
            The seperator for the parameter labels.
        """

        root = f"{root}{self.label}{seperator}" if root is not None else ""
        for label, p in self._parameters.items():
            yield (f"{root}{label}", p)
        for _, l in self.items():
            for (lbl, p) in l.all(root=root, seperator=seperator):
                yield (lbl, p)

    def as_parameter_dict(self) -> Parameters:
        """
        Creates a lmfit.Parameters dictionary from the group.

        Notes
        -----

        Only for internal use.
        """

        params = Parameters()
        for label, p in self.all(seperator="_"):
            p.name = "_" + label
            if p.non_neg:
                p = copy.deepcopy(p)
                if p.min == 1:
                    p.min += 1e-10
                try:
                    p.min = log(p.min) if np.isfinite(p.min) else p.min
                except Exception:
                    raise Exception("Could not take log of minimum of parameter"
                                    f" '{label}' with value '{p.min}'")
                if p.max == 1:
                    p.max += 1e-10
                try:
                    p.max = log(p.max) if np.isfinite(p.max) else p.max
                except Exception:
                    raise Exception("Could not take log of maximum of parameter"
                                    f" '{label}' with value '{p.max}'")
                if p.value == 1:
                    p.value += 1e-10
                try:
                    p.value = log(p.value)
                except Exception:
                    raise Exception("Could not take log of parameter"
                                    f" '{label}' with value '{p.value}'")
            params.add(p)
        return params

    def markdown(self) -> str:
        """Formats the :class:`ParameterGroup` as markdown string."""
        t = "".join(["  " for _ in range(self.get_nr_roots())])
        s = ""
        if self.label != "p":
            s += f"{t}* __{self.label}__:\n"
        for _, p in self._parameters.items():
            s += f"{t}  * {p}\n"
        for _, g in self.items():
            s += f"{g.__str__()}"
        return s

    def __str__(self):
        return self.markdown()



import numpy as np
import xarray as xr

from glotaran.parameter import ParameterGroup
from glotaran.builtin.models.kinetic_spectrum import KineticSpectrumModel
from glotaran.builtin.models.kinetic_spectrum.spectral_relations import \
    create_spectral_relation_matrix
from glotaran.builtin.models.kinetic_image.kinetic_image_matrix \
    import kinetic_image_matrix


def test_spectral_relation():
    model = KineticSpectrumModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1', 's2', 's3', 's4'],
                'parameters': ['i.1', 'i.2', 'i.3', 'i.4']
            },

        },
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {
                ("s1", "s1"): 'kinetic.1',
                ("s2", "s2"): 'kinetic.1',
                ("s3", "s3"): 'kinetic.1',
                ("s4", "s4"): 'kinetic.1',
            }}
        },
        'spectral_relations': [
            {
                'compartment': 's1',
                'target': 's2',
                'parameter': 'rel.1',
                'interval': [(0, 2)],
            },
            {
                'compartment': 's1',
                'target': 's3',
                'parameter': 'rel.2',
                'interval': [(0, 2)],
            },
        ],
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'megacomplex': ['mc1'],
            },
        },
    })
    print(model)

    rel1, rel2 = 10, 20
    parameter = ParameterGroup.from_dict({
        'kinetic': [1e-4],
        'i': [1, 2, 3, 4],
        'rel': [rel1, rel2],
    })

    time = np.asarray(np.arange(0, 50, 1.5))
    dataset = model.dataset['dataset1'].fill(model, parameter)
    compartments, matrix = kinetic_image_matrix(dataset, time, 0)

    assert len(compartments) == 4
    assert matrix.shape == (time.size, 4)

    reduced_compartments, relation_matrix = \
        create_spectral_relation_matrix(model, parameter, compartments, matrix, 1)

    print(relation_matrix)
    assert len(reduced_compartments) == 2
    assert relation_matrix.shape == (4, 2)
    assert np.array_equal(
        relation_matrix,  [
            [1., 0.],
            [10., 0.],
            [20., 0.],
            [0., 1.],
        ]
    )

    reduced_compartments, reduced_matrix = \
        model.constrain_matrix_function(parameter, compartments, matrix, 1)

    assert reduced_matrix.shape == (time.size, 2)

    print(reduced_matrix[0, 0], matrix[0, 0], matrix[0, 1], matrix[0, 2])
    assert np.allclose(
        reduced_matrix[:, 0],  matrix[:, 0] + rel1 * matrix[:, 1] + rel2 * matrix[:, 2]
    )

    clp = xr.DataArray([[1., 10., 20., 1]],
                       coords=(('spectral', [1]), ('clp_label', ['s1', 's2', 's3', 's4'])))

    data = model.simulate('dataset1', parameter, clp=clp,
                          axes={'time': time, 'spectral': np.array([1])})

    result = model.optimize(parameter, {'dataset1': data}, max_nfev=1)

    result_data = result.data['dataset1']
    print(result_data.species_associated_spectra)
    assert result_data.species_associated_spectra.shape == (1, 4)
    assert result_data.species_associated_spectra[0, 1] == \
        rel1 * result_data.species_associated_spectra[0, 0]
    assert result_data.species_associated_spectra[0, 2] == \
        rel2 * result_data.species_associated_spectra[0, 0]


if __name__ == "__main__":
    test_spectral_relation()



""" K-Matrix """

from collections import OrderedDict
import itertools
import typing

import numpy as np
import scipy

from glotaran.model import model_attribute
from glotaran.parameter import Parameter

from .initial_concentration import InitialConcentration


@model_attribute(
    properties={
        'matrix': {'type': typing.Dict[typing.Tuple[str, str], Parameter]},
    },
)
class KMatrix:
    """ A K-Matrix represents a first order differental system."""

    @classmethod
    def empty(cls, label: str, compartments: typing.List[str]) -> 'KMatrix':
        """ Creates an empty K-Matrix. Useful for combining.

        Parameters
        ----------
        label :
            Label of the K-Matrix

        compartments :
            A list of all compartments in the model.
        """
        return cls(label, OrderedDict())

    def involved_compartments(self) -> typing.List[str]:
        """ A list of all compartments in the Matrix. """
        compartments = []
        for index in self.matrix:
            compartments.append(index[0])
            compartments.append(index[1])

        compartments = list(set(compartments))
        return compartments

    def combine(self, k_matrix: "KMatrix") -> "KMatrix":
        """ Creates a combined matrix.

        Parameters
        ----------
        k_matrix :
            KMatrix to combine with.


        Returns
        -------
        combined :
            The combined KMatrix.

        """
        if not isinstance(k_matrix, KMatrix):
            raise TypeError("K-Matrices can only be combined with other K-Matrices.")
        combined_matrix = {}
        for entry in k_matrix.matrix:
            combined_matrix[entry] = k_matrix.matrix[entry]
        for entry in self.matrix:
            combined_matrix[entry] = self.matrix[entry]
        combined = KMatrix()
        combined.label = f"{self.label}+{k_matrix.label}"
        combined.matrix = combined_matrix
        return combined

    def matrix_as_markdown(self, compartments: typing.List[str] = None,
                           fill_parameter: bool = False,
                           ) -> str:
        """ Returns the KMatrix as markdown formatted table.

        Parameters
        ----------
        compartments :
            (default = None)
            An optional list defining the desired order of compartments.
        fill_parameter : bool
            (default = False)
            If true, the entries will be filled with the actual parameter value
            instead of abels.
        """

        compartments = [c for c in compartments if c in self.involved_compartments()] \
            if compartments else self.involved_compartments()
        size = len(compartments)
        array = np.zeros((size, size), dtype=np.object)
        # Matrix is a dict
        for index in self.matrix:
            i = compartments.index(index[0])
            j = compartments.index(index[1])
            array[i, j] = \
                self.matrix[index].full_label if not fill_parameter else self.matrix[index].value
        return self._array_as_markdown(array, compartments, compartments)

    def a_matrix_as_markdown(self, initial_concentration: InitialConcentration) -> str:
        """ Returns the A Matrix as markdown formatted table.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        compartments = [c for c in initial_concentration.compartments
                        if c in self.involved_compartments()]
        return self._array_as_markdown(self.a_matrix(initial_concentration).T,
                                       compartments,
                                       self.rates(initial_concentration),
                                       )

    @staticmethod
    def _array_as_markdown(array, row_header, column_header):
        markdown = "| compartment | "
        markdown += " | ".join([f"{e:.4e}" if not isinstance(e, str)
                                else e for e in column_header])
        markdown += "\n|"
        markdown += "|".join(['---' for _ in range(len(column_header)+1)])
        markdown += "\n"

        for i, row in enumerate(array):
            markdown += f"| {row_header[i]} | " if isinstance(row_header[i], str) \
                else f"| {row_header[i]:.4e} | "
            markdown += " | ".join([f"{e:.4e}" if not isinstance(e, str) else e for e in row])
            markdown += "|\n"

        return markdown

    def reduced(self, compartments: typing.List[str]) -> np.ndarray:
        """ The reduced representation of the KMatrix as numpy array.

        Parameters
        ----------
        compartments :
            The compartment order.
        """

        compartments = [c for c in compartments if c in self.involved_compartments()]
        size = len(compartments)
        array = np.zeros((size, size), dtype=np.float64)
        # Matrix is a dict
        for index in self.matrix:
            i = compartments.index(index[0])
            j = compartments.index(index[1])
            array[i, j] = self.matrix[index]
        return array

    def full(self, compartments: typing.List[str]) -> np.ndarray:
        """ The full representation of the KMatrix as numpy array.

        Parameters
        ----------
        compartments :
            The compartment order.
        """
        compartments = [c for c in compartments if c in self.involved_compartments()]
        size = len(compartments)
        mat = np.zeros((size, size), np.float64)
        for (to_comp, from_comp), param in self.matrix.items():
            to_idx = compartments.index(to_comp)
            fr_idx = compartments.index(from_comp)

            if to_idx == fr_idx:
                mat[to_idx, fr_idx] -= param
            else:
                mat[to_idx, fr_idx] += param
                mat[fr_idx, fr_idx] -= param
        return mat

    def eigen(self, compartments: typing.List[str]) -> typing.Tuple[np.ndarray, np.ndarray]:
        """ Returns the eigenvalues and eigenvectors of the k matrix.

        Parameters
        ----------
        compartments :
            The compartment order.
        """
        # We take the transpose to be consistent with timp
        matrix = self.full(compartments).T
        # get the eigenvectors and values, we take the left ones to have
        # computation consistent with TIMP
        eigenvalues, eigenvectors = scipy.linalg.eig(matrix, left=True,
                                                     right=False)
        return (eigenvalues.real, eigenvectors.real)

    def rates(self, initial_concentration: InitialConcentration) -> np.ndarray:
        """ The resulting rates of the matrix.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        if self.is_unibranched(initial_concentration):
            return np.diag(self.full(initial_concentration.compartments)).copy()
        else:
            rates, _ = self.eigen(initial_concentration.compartments)
            return rates

    def _gamma(self,
               eigenvectors: np.ndarray,
               initial_concentration: InitialConcentration,
               ) -> np.ndarray:
        compartments = [c for c in initial_concentration.compartments
                        if c in self.involved_compartments()]
        initial_concentration = \
            [initial_concentration.parameters[initial_concentration.compartments.index(c)]
             for c in compartments]

        gamma = scipy.linalg.solve(eigenvectors, initial_concentration)
        return np.diag(gamma)

    def a_matrix(self, initial_concentration: InitialConcentration) -> np.ndarray:
        """ The resulting A matrix of the KMatrix.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        return self.a_matrix_unibranch(initial_concentration) \
            if self.is_unibranched(initial_concentration) \
            else self.a_matrix_non_unibranch(initial_concentration)

    def a_matrix_non_unibranch(self, initial_concentration: InitialConcentration) -> np.ndarray:
        """ The resulting A matrix of the KMatrix for a non-unibranched model.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        eigenvalues, eigenvectors = self.eigen(initial_concentration.compartments)
        gamma = self._gamma(eigenvectors, initial_concentration)

        a_matrix = eigenvectors @ gamma

        return a_matrix.T

    def a_matrix_unibranch(self, initial_concentration: InitialConcentration) -> np.array:
        """ The resulting A matrix of the KMatrix for an unibranched model.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        compartments = [c for c in initial_concentration.compartments
                        if c in self.involved_compartments()]
        matrix = self.full(compartments).T
        rates = np.diag(matrix)

        a_matrix = np.zeros(matrix.shape, dtype=np.float64)
        a_matrix[0, 0] = 1.0
        for i, j in itertools.product(range(rates.size), range(1, rates.size)):
            if i > j:
                continue
            a_matrix[i, j] = np.prod([rates[m] for m in range(j)]) / \
                np.prod([rates[m] - rates[i] for m in range(j+1) if not i == m])
        return a_matrix

    def is_unibranched(self, initial_concentration: InitialConcentration) -> bool:
        """ Returns true in the KMatrix represents an unibranched model.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        if not np.sum(
            [initial_concentration.parameters[initial_concentration.compartments.index(c)]
             for c in self.involved_compartments()]) == 1:
            return False
        matrix = self.reduced(initial_concentration.compartments)
        for i in range(matrix.shape[1]):
            if not np.nonzero(matrix[:, i])[0].size == 1 or i != 0 and matrix[i, i-1] == 0:
                return False
        return True



"""Glotaran DOAS Model Oscillation"""

from glotaran.model import model_attribute
from glotaran.parameter import Parameter


@model_attribute(
    properties={
        'frequency': Parameter,
        'rate': Parameter,
    })
class Oscillation:
    """A damped oscillation"""



import numpy as np
import glotaran as gta

sim_model = gta.builtin.models.kinetic_spectrum.KineticSpectrumModel.from_dict({
    'initial_concentration': {
        'j1': {
            'compartments': ['s1', 's2', 's3'],
            'parameters': ['j.1', 'j.0', 'j.0'],
        },
    },
    'k_matrix': {
        "k1": {'matrix': {
            ("s2", "s1"): 'kinetic.1',
            ("s3", "s2"): 'kinetic.2',
            ("s3", "s3"): 'kinetic.3',
        }}
    },

    'megacomplex': {
        'm1': {
           'k_matrix': ['k1'],
        }
    },
    'shape': {
        'sh1': {
            'type': "gaussian",
            'amplitude': "shape.amps.1",
            'location': "shape.locs.1",
            'width': "shape.width.1",
        },
        'sh2': {
            'type': "gaussian",
            'amplitude': "shape.amps.2",
            'location': "shape.locs.2",
            'width': "shape.width.2",
        },
        'sh3': {
            'type': "gaussian",
            'amplitude': "shape.amps.3",
            'location': "shape.locs.3",
            'width': "shape.width.3",
        },
    },
    'irf': {
        'irf1': {'type': 'gaussian', 'center': 'irf.center', 'width': 'irf.width'},
    },
    'dataset': {
        'dataset1': {
            'initial_concentration': 'j1',
            'megacomplex': ['m1'],
            'shape': {
                's1': 'sh1',
                's2': 'sh2',
                's3': 'sh3',
            },
            'irf': 'irf1',
        }
    }
    })

wanted_parameter = gta.ParameterGroup.from_dict({
    'j': [
            ['1', 1, {'non-negative': False, 'vary': False}],
            ['0', 0, {'non-negative': False, 'vary': False}],
        ],
    'kinetic': [
        ["1", 0.5],
        ["2", 0.3],
        ["3", 0.1],
    ],
    'osc': [
        ['freq', 25],
        ['rate', 1],
    ],
    'shape': {'amps': [30, 20, 40], 'locs': [620, 630, 650], 'width': [40, 20, 60]},
    'irf': [['center', 0.3], ['width', 0.1]],
    })

parameter = gta.ParameterGroup.from_dict({
    'j': [
            ['1', 1, {'vary': False, 'non-negative': False}],
            ['0', 0, {'vary': False, 'non-negative': False}],
        ],
    'kinetic': [
        ["1", 0.5],
        ["2", 0.3],
        ["3", 0.1],
    ],
    'irf': [['center', 0.3], ['width', 0.1]],
    })

_time = np.arange(-1, 20, 0.01)
_spectral = np.arange(600, 700, 1.4)

dataset = sim_model.simulate('dataset1', wanted_parameter,
                             {'time': _time, 'spectral': _spectral},
                             noise=True, noise_std_dev=1e-2)

model = gta.builtin.models.kinetic_spectrum.KineticSpectrumModel.from_dict({
    'initial_concentration': {
        'j1': {
                'compartments': ['s1', 's2', 's3'],
                'parameters': ['j.1', 'j.0', 'j.0']
              },
    },
    'k_matrix': {
        "k1": {'matrix': {
            ("s2", "s1"): 'kinetic.1',
            ("s3", "s2"): 'kinetic.2',
            ("s3", "s3"): 'kinetic.3',
        }}
    },

    'megacomplex': {
        'm1': {
           'k_matrix': ['k1'],
        }
    },
    'irf': {
        'irf1': {'type': 'gaussian', 'center': 'irf.center', 'width': 'irf.width'},
    },
    'dataset': {
        'dataset1': {
            'initial_concentration': 'j1',
            'megacomplex': ['m1'],
            'irf': 'irf1',
        }
    }
})



import numpy as np
import xarray as xr

from glotaran.analysis.scheme import Scheme
from glotaran.parameter import ParameterGroup

from glotaran.analysis.problem_bag import create_grouped_bag


from .mock import MockModel


def test_single_dataset():
    model = MockModel.from_dict({
        "dataset": {
            "dataset1": {
                "megacomplex": [],
            },
        }
    })
    print(model.validate())
    assert model.valid()

    parameter = ParameterGroup.from_list([1, 10])
    print(model.validate(parameter))
    assert model.valid(parameter)
    axis_e = [1, 2, 3]
    axis_c = [5, 7, 9, 12]

    data = {'dataset1': xr.DataArray(
        np.ones((3, 4)),
        coords=[('e', axis_e), ('c', axis_c)]
    ).to_dataset(name="data")}

    scheme = Scheme(model, parameter, data)
    bag, datasets = create_grouped_bag(scheme)
    bag = bag.compute()
    assert len(datasets) == 0
    assert len(bag) == 3
    assert all([p.data.size == 4 for p in bag])
    assert all([p.descriptor[0].dataset == 'dataset1' for p in bag])
    assert all([all(p.descriptor[0].axis == axis_c) for p in bag])
    assert [p.descriptor[0].index for p in bag] == axis_e


def test_multi_dataset_no_overlap():
    model = MockModel.from_dict({
        "dataset": {
            "dataset1": {
                "megacomplex": [],
            },
            "dataset2": {
                "megacomplex": [],
            },
        }
    })

    print(model.validate())
    assert model.valid()

    parameter = ParameterGroup.from_list([1, 10])
    print(model.validate(parameter))
    assert model.valid(parameter)

    axis_e_1 = [1, 2, 3]
    axis_c_1 = [5, 7]
    axis_e_2 = [4, 5, 6]
    axis_c_2 = [5, 7, 9]
    data = {
        'dataset1': xr.DataArray(
            np.ones((3, 2)),
            coords=[('e', axis_e_1), ('c', axis_c_1)]
        ).to_dataset(name="data"),
        'dataset2': xr.DataArray(
            np.ones((3, 3)),
            coords=[('e', axis_e_2), ('c', axis_c_2)]
        ).to_dataset(name="data"),
    }

    scheme = Scheme(model, parameter, data)
    bag, datasets = create_grouped_bag(scheme)
    bag = bag.compute()
    assert len(datasets) == 0
    assert len(bag) == 6
    assert all([p.data.size == 2 for p in bag[:3]])
    assert all([p.descriptor[0].dataset == 'dataset1' for p in bag[:3]])
    assert all([all(p.descriptor[0].axis == axis_c_1) for p in bag[:3]])
    assert [p.descriptor[0].index for p in bag[:3]] == axis_e_1

    assert all([p.data.size == 3 for p in bag[3:]])
    assert all([p.descriptor[0].dataset == 'dataset2' for p in bag[3:]])
    assert all([all(p.descriptor[0].axis == axis_c_2) for p in bag[3:]])
    assert [p.descriptor[0].index for p in bag[3:]] == axis_e_2


def test_multi_dataset_overlap():
    model = MockModel.from_dict({
        "dataset": {
            "dataset1": {
                "megacomplex": [],
            },
            "dataset2": {
                "megacomplex": [],
            },
        }
    })

    print(model.validate())
    assert model.valid()

    parameter = ParameterGroup.from_list([1, 10])
    print(model.validate(parameter))
    assert model.valid(parameter)

    axis_e_1 = [0, 1, 2, 3]
    axis_c_1 = [5, 7]
    axis_e_2 = [1.4, 2.4, 3.4, 9]
    axis_c_2 = [5, 7, 9, 12]
    data = {
        'dataset1': xr.DataArray(
            np.ones((4, 2)),
            coords=[('e', axis_e_1), ('c', axis_c_1)]
        ).to_dataset(name="data"),
        'dataset2': xr.DataArray(
            np.ones((4, 4)),
            coords=[('e', axis_e_2), ('c', axis_c_2)]
        ).to_dataset(name="data"),
    }

    scheme = Scheme(model, parameter, data, group_tolerance=5e-1)
    bag, datasets = create_grouped_bag(scheme)
    bag = bag.compute()
    assert len(datasets) == 1
    assert "dataset1dataset2" in datasets
    assert datasets['dataset1dataset2'] == ["dataset1", "dataset2"]
    assert len(bag) == 5

    assert all([p.data.size == 2 for p in bag[:1]])
    assert all([p.descriptor[0].dataset == 'dataset1' for p in bag[:4]])
    assert all([all(p.descriptor[0].axis == axis_c_1) for p in bag[:4]])
    assert [p.descriptor[0].index for p in bag[:4]] == axis_e_1

    assert all([p.data.size == 6 for p in bag[1:4]])
    assert all([p.descriptor[1].dataset == 'dataset2' for p in bag[1:4]])
    assert all([all(p.descriptor[1].axis == axis_c_2) for p in bag[1:4]])
    assert [p.descriptor[1].index for p in bag[1:4]] == axis_e_2[:3]

    assert all([p.data.size == 4 for p in bag[4:]])
    assert all([p.descriptor[0].dataset == 'dataset2' for p in bag[4:]])
    assert all([all(p.descriptor[0].axis == axis_c_2) for p in bag[4:]])
    assert [p.descriptor[0].index for p in bag[4:]] == axis_e_2[3:]



import sys

import click
from click import echo, prompt

import glotaran as gta


def signature_analysis(cmd):
    cmd = click.option('--model', '-m', default=None, type=click.Path(exists=True, dir_okay=False),
                       help='Path to model file.')(cmd)
    cmd = click.option('--parameter', '-p',
                       default=None,
                       type=click.Path(exists=True, dir_okay=False),
                       help='(optional) Path to parameter file.'
                       )(cmd)
    cmd = click.argument("scheme",
                         type=click.Path(exists=True, dir_okay=False),
                         required=False
                         )(cmd)
    return cmd


def _load_file(filename, loader, name, verbose):
    try:
        if verbose:
            echo(f"Loading {name} file '{filename}'... ", nl=False)
        result = loader(filename)
        if verbose:
            echo("Success")
        return result
    except Exception as e:
        if verbose:
            echo(message="Error", err=True)
        else:
            echo(message=f"Error parsing {name} file: \n", err=True)
        echo(message=e, err=True)
        sys.exit(1)


def load_scheme_file(filename, verbose=False):
    return _load_file(filename, gta.analysis.scheme.Scheme.from_yml_file, 'scheme', verbose)


def load_model_file(filename, verbose=False):
    return _load_file(filename, gta.read_model_from_yml_file, 'model', verbose)


def load_parameter_file(filename, fmt=None, verbose=False):

    def loader(filename):
        return gta.parameter.ParameterGroup.from_file(filename, fmt=fmt)

    return _load_file(filename, loader, 'parameter', verbose)


def load_dataset_file(filename, fmt=None, verbose=False):

    def loader(filename):
        return gta.io.read_data_file(filename, fmt=fmt)

    return _load_file(filename, loader, 'parameter', verbose)


def select_name(filename, dataset):

    names = [n for n in dataset]
    echo(f"\nDataset names in in '{filename}':\n")
    for i, n in enumerate(names):
        echo(f"* [{i}] {n}")

    n = prompt("\n Please select a name to export",
               type=click.IntRange(min=0, max=len(names)-1),
               show_choices=True,
               )
    return names[n]


def select_data(data, dim, selection):
    try:
        [float(i) for i in data.coords[dim].values]
        numeric = True
    except ValueError:
        numeric = False
    method = None
    if numeric:
        try:
            if isinstance(selection, tuple):
                selection = tuple(float(c) for c in selection)
            elif isinstance(selection, list):
                selection = [float(c) for c in selection]
            else:
                selection = float(selection)
        except ValueError:
            raise ValueError(f"Error: Selection '{selection}' is not numeric")
        method = 'nearest'
    if isinstance(selection, tuple):
        min = selection[0]
        max = selection[1]
        selection = data.coords[dim]
        selection = selection.where(selection >= min, drop=True)
        selection = selection.where(selection <= max, drop=True)
    return data.sel({dim: selection}, method=method)


def write_data(data, out):
    df = data.to_dataframe()
    if len(data.dims) == 2:
        df = df.reset_index().pivot(index=data.dims[0], columns=data.dims[1], values=data.name)
    df.to_csv(out)


class ValOrRangeOrList(click.ParamType):
    name = 'number or range or list'

    def convert(self, value, param, ctx):
        if value[0] == '(' and value[-1] == ')':
            split = value[1:-1].split(',')
            if not len(split) == 2:
                self.fail(f"Malformed range: '{value}'")
            return (split[0].strip(), split[1].strip())
        if value[0] == '[' and value[-1] == ']':
            split = value[1:-1].split(',')
            return [s.strip() for s in split]
        return value


VALORRANGEORLIST = ValOrRangeOrList()



"""A base class for global analysis models."""


import copy
import inspect
import typing
import numpy as np
import xarray as xr

from glotaran.analysis.result import Result
from glotaran.analysis.scheme import Scheme
from glotaran.analysis.simulation import simulate
from glotaran.analysis.optimize import optimize
from glotaran.parameter import ParameterGroup


class Model:
    """A base class for global analysis models."""

    @classmethod
    def from_dict(cls, model_dict_ref: typing.Dict) -> typing.Type['Model']:
        """Creates a model from a dictionary.

        Parameters
        ----------
        model_dict :
            Dictionary containing the model.
        """

        model = cls()

        model_dict = copy.deepcopy(model_dict_ref)

        # iterate over items
        for name, attribute in list(model_dict.items()):

            # we determine if we the item is known by the model by looking for
            # a setter with same name.

            if hasattr(model, f'set_{name}'):

                # get the set function
                set = getattr(model, f'set_{name}')

                # we retrieve the actual class from the signature
                for label, item in attribute.items():
                    item_cls = set.__func__.__annotations__['item']
                    is_typed = hasattr(item_cls, "_glotaran_model_attribute_typed")
                    if isinstance(item, dict):
                        if is_typed:
                            if 'type' not in item:
                                raise Exception(f"Missing type for attribute '{name}'")
                            item_type = item['type']

                            if item_type not in item_cls._glotaran_model_attribute_types:
                                raise Exception(f"Unknown type '{item_type}' "
                                                f"for attribute '{name}'")
                            item_cls = \
                                item_cls._glotaran_model_attribute_types[item_type]
                        item['label'] = label
                        set(label, item_cls.from_dict(item))
                    elif isinstance(item, list):
                        if is_typed:
                            if len(item) < 2 and len(item) != 1:
                                raise Exception(f"Missing type for attribute '{name}'")
                            item_type = item[1] if len(item) != 1 and \
                                hasattr(item_cls, 'label') else item[0]

                            if item_type not in item_cls._glotaran_model_attribute_types:
                                raise Exception(f"Unknown type '{item_type}' "
                                                f"for attribute '{name}'")
                            item_cls = \
                                item_cls._glotaran_model_attribute_types[item_type]
                        item = [label] + item
                        set(label, item_cls.from_list(item))
                del model_dict[name]

            elif hasattr(model, f'add_{name}'):

                # get the set function
                add = getattr(model, f'add_{name}')

                # we retrieve the actual class from the signature
                for item in attribute:
                    item_cls = add.__func__.__annotations__['item']
                    is_typed = hasattr(item_cls, "_glotaran_model_attribute_typed")
                    if isinstance(item, dict):
                        if is_typed:
                            if 'type' not in item:
                                raise Exception(f"Missing type for attribute '{name}'")
                            item_type = item['type']

                            if item_type not in item_cls._glotaran_model_attribute_types:
                                raise Exception(f"Unknown type '{item_type}' "
                                                f"for attribute '{name}'")
                            item_cls = \
                                item_cls._glotaran_model_attribute_types[item_type]
                        add(item_cls.from_dict(item))
                    elif isinstance(item, list):
                        if is_typed:
                            if len(item) < 2 and len(item) != 1:
                                raise Exception(f"Missing type for attribute '{name}'")
                            item_type = item[1] if len(item) != 1 and \
                                hasattr(item_cls, 'label') else item[0]

                            if item_type not in item_cls._glotaran_model_attribute_types:
                                raise Exception(f"Unknown type '{item_type}' "
                                                f"for attribute '{name}'")
                            item_cls = \
                                item_cls._glotaran_model_attribute_types[item_type]
                        add(item_cls.from_list(item))
                del model_dict[name]

        return model

    @property
    def index_depended_matrix(self):
        return len(inspect.signature(self.matrix).parameters) == 3

    @property
    def model_type(self) -> str:
        """The type of the model as human readable string."""
        return self._model_type

    def simulate(self,
                 dataset: str,
                 parameter: ParameterGroup,
                 axes: typing.Dict[str, np.ndarray] = None,
                 clp: typing.Union[np.ndarray, xr.DataArray] = None,
                 noise: bool = False,
                 noise_std_dev: float = 1.0,
                 noise_seed: int = None,
                 ) -> xr.Dataset:
        """Simulates the model.

        Parameters
        ----------
        dataset :
            Label of the dataset to simulate.
        parameter :
            The parameters for the simulation.
        axes :
            A dictory with axes for simulation.
        clp :
            Conditionaly linear parameter. Will be used instead of `model.global_matrix` if given.
        noise :
            If `True` noise is added to the simulated data.
        noise_std_dev :
            The standart deviation of the noise.
        noise_seed :
            Seed for the noise.
        """
        return simulate(self, dataset, parameter, axes=axes, clp=clp, noise=noise,
                        noise_std_dev=noise_std_dev, noise_seed=noise_seed)

    def optimize(self,
                 parameter: ParameterGroup,
                 data: typing.Dict[str, typing.Union[xr.Dataset, xr.DataArray]],
                 nnls: bool = False,
                 verbose: bool = True,
                 max_nfev: int = None,
                 group_tolerance: int = 0,
                 client=None,
                 ) -> Result:
        """Optimizes the parameter for this model.

        Parameters
        ----------
        data :
            A dictonary containing all datasets with their labels as keys.
        parameter : glotaran.model.ParameterGroup
            The initial parameter.
        nnls :
            If `True` non-linear least squaes optimizing is used instead of variable projection.
        verbose :
            If `True` feedback is printed at every iteration.
        max_nfev :
            Maximum number of function evaluations. `None` for unlimited.
        group_tolerance :
            The tolerance for grouping datasets along the global dimension.
        """
        scheme = Scheme(model=self, parameter=parameter, data=data,
                        nnls=nnls, group_tolerance=group_tolerance, nfev=max_nfev)
        result = optimize(scheme, verbose=verbose, client=client)
        return result

    def result_from_parameter(self,
                              parameter: ParameterGroup,
                              data: typing.Dict[str, typing.Union[xr.DataArray, xr.Dataset]],
                              nnls: bool = False, group_atol: float = 0.0
                              ) -> Result:
        """Loads a result from parameters without optimization.

        Parameters
        ----------
        data :
            A dictonary containing all datasets with their labels as keys.
        parameter : glotaran.model.ParameterGroup
            The parameter.
        nnls :
            If `True` non-linear least squaes optimizing is used instead of variable projection.
        group_atol :
            The tolerance for grouping datasets along the global axes.

        """
        return Result.from_parameter(self, data, parameter, nnls, group_atol)

    def problem_list(self, parameter: ParameterGroup = None) -> typing.List[str]:
        """
        Returns a list with all problems in the model and missing parameters if specified.

        Parameters
        ----------

        parameter :
            The parameter to validate.
        """
        problems = []

        attrs = getattr(self, '_glotaran_model_attributes')
        for attr in attrs:
            attr = getattr(self, attr)
            if isinstance(attr, list):
                for item in attr:
                    problems += item.validate(self, parameter=parameter)
            else:
                for _, item in attr.items():
                    problems += item.validate(self, parameter=parameter)

        return problems

    def validate(self, parameter: ParameterGroup = None) -> str:
        """
        Returns a string listing all problems in the model and missing parameters if specified.

        Parameters
        ----------

        parameter :
            The parameter to validate.
        """
        result = ""

        problems = self.problem_list(parameter)
        if problems:
            result = f"Your model has {len(problems)} problems:\n"
            for p in problems:
                result += f"\n * {p}"
        else:
            result = "Your model is valid."
        return result

    def valid(self, parameter: ParameterGroup = None) -> bool:
        """Returns `True` if the number problems in the model is 0, else `False`

        Parameters
        ----------

        parameter :
            The parameter to validate.
        """
        return len(self.problem_list(parameter)) == 0

    def markdown(self, parameter: ParameterGroup = None, initial: ParameterGroup = None) -> str:
        """Formats the model as Markdown string.

        Parameters will be included if specified.

        Parameters
        ----------
        parameter :
            Parameter to include.
        initial :
            Initial values for the parameters.
        """
        attrs = getattr(self, '_glotaran_model_attributes')
        string = "# Model\n\n"
        string += f"_Type_: {self.model_type}\n\n"

        for attr in attrs:
            items = getattr(self, attr)
            if not items:
                continue

            string += f"## {attr.replace('_', ' ').title()}\n"
            string += "\n"

            if isinstance(items, dict):
                items = items.values()
            for item in items:
                item_str = item.mprint(parameter=parameter, initial=initial).split('\n')
                string += f'* {item_str[0]}\n'
                for s in item_str[1:]:
                    string += f"  {s}\n"
            string += "\n"
        return string

    def __str__(self):
        return self.markdown()



"""Glotaran kinetic spectrum Matrix"""

import numpy as np

from glotaran.builtin.models.kinetic_image.kinetic_image_matrix import \
    kinetic_image_matrix

from .spectral_irf import IrfGaussianCoherentArtifact


def kinetic_spectrum_matrix(dataset_descriptor=None, axis=None, index=None, irf=None):

    clp_label, matrix = kinetic_image_matrix(
        dataset_descriptor, axis, index, dataset_descriptor.irf)

    if isinstance(dataset_descriptor.irf, IrfGaussianCoherentArtifact):
        irf_clp_label, irf_matrix = \
            dataset_descriptor.irf.calculate_coherent_artifact(axis)
        if matrix is None:
            clp_label = irf_clp_label
            matrix = irf_matrix
        else:
            clp_label += irf_clp_label
            matrix = np.concatenate((matrix, irf_matrix), axis=1)

    return (clp_label, matrix)



""" Glotaran DOAS Model Megacomplex """

from typing import List

from glotaran.model import model_attribute
from glotaran.builtin.models.kinetic_image.kinetic_image_megacomplex \
    import KineticImageMegacomplex


@model_attribute(properties={'oscillation': {'type': List[str], 'default': []}})
class DOASMegacomplex(KineticImageMegacomplex):
    """A Megacomplex with one or more oscillations."""



"""This package contains the intial concentration item."""

import copy
import typing
import numpy as np

from glotaran.model import model_attribute, DatasetDescriptor
from glotaran.parameter import Parameter


@model_attribute(
    properties={
        'compartments': typing.List[str],
        'parameters': typing.List[Parameter],
        'exclude_from_normalize': {'type': typing.List[str], 'default': []},
    }
)
class InitialConcentration:
    """An initial concentration describes the population of the compartments at
    the beginning of an experiment."""

    def normalized(self, dataset: DatasetDescriptor) -> 'InitialConcentration':
        parameters = self.parameters
        for megacomplex in dataset.megacomplex:
            scale = [megacomplex.scale if c in megacomplex.involved_compartments and
                     megacomplex.scale else 1 for c in self.compartments]
            parameters = np.multiply(parameters, scale)
        idx = [c not in self.exclude_from_normalize for c in self.compartments]
        parameters[idx] /= np.sum(parameters[idx])
        new = copy.deepcopy(self)
        for i, value in enumerate(parameters):
            new.parameters[i].value = value
        return new



import pytest
import xarray as xr

from glotaran.analysis.scheme import Scheme
from .mock import MockModel # noqa


@pytest.fixture(scope='session')
def scheme(tmpdir_factory):

    path = tmpdir_factory.mktemp("scheme")

    model = "type: mock\ndataset:\n  dataset1:\n    megacomplex: []"
    model_path = path.join("model.yml")
    with open(model_path, 'w') as f:
        f.write(model)

    parameter = "[1.0, 67.0]"
    parameter_path = path.join("parameter.yml")
    with open(parameter_path, 'w') as f:
        f.write(parameter)

    dataset_path = path.join('dataset.nc')
    xr.DataArray([1, 2, 3]).to_dataset(name='data').to_netcdf(dataset_path)

    scheme = f"""
    model: {model_path}
    parameter: {parameter_path}
    nnls: True
    nfev: 42
    data:
      dataset1: {dataset_path}
    """
    scheme_path = path.join('scheme.gta')
    with open(scheme_path, 'w') as f:
        f.write(scheme)
    return scheme_path


def test_scheme(scheme):
    scheme = Scheme.from_yml_file(scheme)
    assert scheme.model is not None
    assert scheme.model.model_type == "mock"

    assert scheme.parameter is not None
    assert scheme.parameter.get("1") == 1.0
    assert scheme.parameter.get("2") == 67.0

    assert scheme.nnls
    assert scheme.nfev == 42

    assert 'dataset1' in scheme.data
    assert scheme.data['dataset1'].data.size == 3



from typing import Dict, List, Tuple
import pytest

from glotaran.model import (
    Model,
    model,
    model_attribute,
)
from glotaran.parameter import Parameter, ParameterGroup


@model_attribute(
    properties={
        'param': Parameter,
        'megacomplex': str,
        'param_list': List[Parameter],
        'default_item': {'type': int, 'default': 42},
        'complex': {'type': Dict[Tuple[str, str], Parameter]},
    },
)
class MockAttr:
    pass


@model_attribute()
class MockMegacomplex:
    pass


def matrix_fun(dataset, axis):
    pass


def matrix_fun_index(dataset, axis, index):
    pass


@model('mock', attributes={"test": MockAttr}, megacomplex_type=MockMegacomplex, matrix=matrix_fun)
class MockModel(Model):
    pass


@pytest.fixture
def model():
    d = {
        "megacomplex": {"m1": [], "m2": []},
        "test": {
            "t1": {'param': "foo",
                   'megacomplex': "m1",
                   'param_list': ["bar", "baz"],
                   'complex': {('s1', 's2'): "baz"},
                   },
            "t2": ['baz', 'm2', ['foo'], 7, {}],
        },
        "dataset": {
            "dataset1": {
                "megacomplex": ['m1', 'm2'],
                "scale": "scale_1",
            },
            "dataset2": [['m2'], 'scale_2']
        }
    }
    return MockModel.from_dict(d)


@pytest.fixture
def model_error():
    d = {
        "megacomplex": {"m1": [], "m2": []},
        "test": {
            "t1": {'param': "fool",
                   'megacomplex': "mX",
                   'param_list': ["bar", "bay"],
                   'complex': {('s1', 's3'): "boz"},
                   },
        },
        "dataset": {
            "dataset1": {
                "megacomplex": ['N1', 'N2'],
                "scale": "scale_1",
            },
            "dataset2": [['mrX'], 'scale_3']
        }
    }
    return MockModel.from_dict(d)


@pytest.fixture
def parameter():
    params = [1, 2,
              ['foo', 3],
              ['bar', 4],
              ['baz', 2],
              ['scale_1', 2],
              ['scale_2', 8],
              4e2
              ]
    return ParameterGroup.from_list(params)


def test_model_misc(model):
    assert model.model_type == 'mock'
    assert not model.index_depended_matrix
    model.matrix = matrix_fun_index
    assert model.index_depended_matrix


@pytest.mark.parametrize(
    "attr",
    ["dataset",
     "megacomplex",
     "test"])
def test_model_attr(model, attr):
    assert hasattr(model, attr)
    assert hasattr(model, f'get_{attr}')
    assert hasattr(model, f'set_{attr}')


def test_model_validity(model, model_error, parameter):
    print(model.test['t1'])
    print(model.problem_list())
    print(model.problem_list(parameter))
    assert model.valid()
    assert model.valid(parameter)
    print(model_error.problem_list())
    print(model_error.problem_list(parameter))
    assert not model_error.valid()
    assert len(model_error.problem_list()) == 4
    assert not model_error.valid(parameter)
    assert len(model_error.problem_list(parameter)) == 8


def test_items(model):

    assert 'm1' in model.megacomplex
    assert 'm2' in model.megacomplex

    assert 't1' in model.test
    t = model.get_test('t1')
    assert t.param.full_label == "foo"
    assert t.megacomplex == 'm1'
    assert [p.full_label for p in t.param_list] == ["bar", "baz"]
    assert t.default_item == 42
    assert ('s1', 's2') in t.complex
    assert t.complex[('s1', 's2')].full_label == "baz"
    assert 't2' in model.test
    t = model.get_test('t2')
    assert t.param.full_label == "baz"
    assert t.megacomplex == 'm2'
    assert [p.full_label for p in t.param_list] == ["foo"]
    assert t.default_item == 7
    assert t.complex == {}

    assert 'dataset1' in model.dataset
    assert model.get_dataset('dataset1').megacomplex == ['m1', 'm2']
    assert model.get_dataset('dataset1').scale.full_label == 'scale_1'

    assert 'dataset2' in model.dataset
    assert model.get_dataset('dataset2').megacomplex == ['m2']
    assert model.get_dataset('dataset2').scale.full_label == 'scale_2'


def test_fill(model, parameter):
    dataset = model.get_dataset('dataset1').fill(model, parameter)
    assert [cmplx.label for cmplx in dataset.megacomplex] == ['m1', 'm2']
    assert dataset.scale == 2

    dataset = model.get_dataset('dataset2').fill(model, parameter)
    assert [cmplx.label for cmplx in dataset.megacomplex] == ['m2']
    assert dataset.scale == 8

    t = model.get_test('t1').fill(model, parameter)
    assert t.param == 3
    assert t.megacomplex.label == 'm1'
    assert t.param_list == [4, 2]
    assert t.default_item == 42
    assert t.complex == {('s1', 's2'): 2}
    t = model.get_test('t2').fill(model, parameter)
    assert t.param == 2
    assert t.megacomplex.label == 'm2'
    assert t.param_list == [3]
    assert t.default_item == 7
    assert t.complex == {}



""" Kinetic Image Dataset Descriptor"""

from glotaran.model import DatasetDescriptor, model_attribute


@model_attribute(properties={
    'initial_concentration': {'type': str, 'allow_none': True},
    'irf': {'type': str, 'allow_none': True},
    'baseline': {'type': bool, 'allow_none': True},
})
class KineticImageDatasetDescriptor(DatasetDescriptor):

    def get_k_matrices(self):
        return [mat for mat in [cmplx.full_k_matrix() for cmplx in self.megacomplex] if mat]

    def compartments(self):
        compartments = []
        for k in self.get_k_matrices():
            compartments += k.involved_compartments()
        return list(set(compartments))



from setuptools import setup, find_packages

install_requires = [
    'click>=7.0',
    'cloudpickle>=1.2',  # Needed by dask
    'dask[bag]>=2.8',
    'lmfit>=0.9.13',
    'netCDF4>=1.5.3',
    'numba>=0.48',
    'numpy>=1.17.3',
    'pandas>=0.25.2',
    'pyyaml>=5.2',
    'scipy>=1.3.2',
    'setuptools>=41.0',
    'xarray>=0.14',
]


with open("README.md", "r") as fh:
    long_description = fh.read()

entry_points = """
    [console_scripts]
    glotaran=glotaran.cli.main:glotaran

    [glotaran.plugins]
    kinetic_image_model = glotaran.builtin.models.kinetic_image
    kinetic_spectrum_model = glotaran.builtin.models.kinetic_spectrum
    doas_model = glotaran.builtin.models.doas

    ascii_file = glotaran.builtin.file_formats.ascii
    sdt_file = glotaran.builtin.file_formats.sdt
"""

setup(
    name="glotaran",
    version='0.0.16',
    description='The Glotaran fitting engine.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/glotaran/pyglotaran',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        "Operating System :: OS Independent",
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Chemistry',
    ],
    author='Joern Weissenborn, Joris Snellenburg, Ivo van Stokkum',
    author_email="""joern.weissenborn@gmail.com,
                    j.snellenburg@gmail.com,
                    i.h.m.van.stokkum@vu.nl """,
    license='GPLv3',
    project_urls={
        'GloTarAn Ecosystem': 'http://glotaran.org',
        'Documentation': 'https://glotaran.readthedocs.io',
        'Source': 'https://github.com/glotaran/pyglotaran',
        'Tracker': 'https://github.com/glotaran/pyglotaran/issues',
    },
    python_requires=">=3.6, <3.9",
    packages=find_packages(),
    install_requires=install_requires,
    entry_points=entry_points,
    test_suite='glotaran',
    tests_require=['pytest'],
    zip_safe=True,
)



# Glotaran package __init__.py

from . import model, parameter, io  # noqa: F401

__version__ = '0.0.16'

ParameterGroup = parameter.ParameterGroup

read_parameter_from_csv_file = ParameterGroup.from_csv
read_parameter_from_yml = ParameterGroup.from_yaml
read_parameter_from_yml_file = ParameterGroup.from_yaml_file


from .parse import parser  # noqa: E402

read_model_from_yml = parser.load_yml
read_model_from_yml_file = parser.load_yml_file

import pkg_resources  # noqa: E402

for entry_point in pkg_resources.iter_entry_points('glotaran.plugins'):
    entry_point.load()



import pytest
from typing import List
import numpy as np

from glotaran.analysis.simulation import simulate
from glotaran.analysis.scheme import Scheme
from glotaran.analysis.optimize import optimize

from glotaran.model import DatasetDescriptor, Model, model_attribute, model
from glotaran.parameter import Parameter, ParameterGroup

from .mock import MockMegacomplex


def calculate_kinetic(dataset_descriptor=None, axis=None, index=None, extra_stuff=None):
    kinpar = -1 * np.array(dataset_descriptor.kinetic)
    compartments = [f's{i+1}' for i in range(len(kinpar))]
    array = np.exp(np.outer(axis, kinpar))
    return (compartments, array)


def calculate_spectral_simple(dataset, axis):
    kinpar = -1 * np.array(dataset.kinetic)
    compartments = [f's{i+1}' for i in range(len(kinpar))]
    array = np.asarray([[1 for _ in range(axis.size)] for _ in compartments])
    return compartments, array.T


def calculate_spectral_gauss(dataset, axis):
    location = np.asarray(dataset.location)
    amp = np.asarray(dataset.amplitude)
    delta = np.asarray(dataset.delta)

    array = np.empty((location.size, axis.size), dtype=np.float64)

    for i in range(location.size):
        array[i, :] = amp[i] * np.exp(
            -np.log(2) * np.square(
                2 * (axis - location[i])/delta[i]
            )
        )
    compartments = [f's{i+1}' for i in range(location.size)]
    return compartments, array.T


@model_attribute(properties={
    'kinetic': List[Parameter],
})
class DecayDatasetDescriptor(DatasetDescriptor):
    pass


@model_attribute(properties={
    'kinetic': List[Parameter],
    'location': List[Parameter],
    'amplitude': List[Parameter],
    'delta': List[Parameter],
})
class GaussianShapeDecayDatasetDescriptor(DatasetDescriptor):
    pass


@model('one_channel',
       dataset_type=DecayDatasetDescriptor,
       matrix=calculate_kinetic,
       matrix_dimension='c',
       global_matrix=calculate_spectral_simple,
       global_dimension='e',
       megacomplex_type=MockMegacomplex,
       additional_penalty_function=lambda model, parameter, clp_labels, clps, index: [],
       )
class DecayModel(Model):
    pass


@model('multi_channel',
       dataset_type=GaussianShapeDecayDatasetDescriptor,
       matrix=calculate_kinetic,
       matrix_dimension='c',
       global_matrix=calculate_spectral_gauss,
       global_dimension='e',
       megacomplex_type=MockMegacomplex,
       )
class GaussianDecayModel(Model):
    pass


class OneCompartmentDecay:
    wanted = ParameterGroup.from_list([101e-4])
    initial = ParameterGroup.from_list([100e-5])

    e_axis = np.asarray([1])
    c_axis = np.arange(0, 150, 1.5)

    model = DecayModel.from_dict({
        'compartment': ["s1"],
        'dataset': {
            'dataset1': {
                'initial_concentration': [],
                'megacomplex': [],
                'kinetic': ['1']
            }
        }
    })
    sim_model = model


class TwoCompartmentDecay:
    wanted = ParameterGroup.from_list([11e-4, 22e-5])
    initial = ParameterGroup.from_list([10e-4, 20e-5])

    e_axis = np.asarray([1])
    c_axis = np.arange(0, 150, 1.5)

    model = DecayModel.from_dict({
        'compartment': ["s1", "s2"],
        'dataset': {
            'dataset1': {
                'initial_concentration': [],
                'megacomplex': [],
                'kinetic': ['1', '2']
            }
        }
    })
    sim_model = model


class MultichannelMulticomponentDecay:
    wanted = ParameterGroup.from_dict({
        'k': [.006, 0.003, 0.0003, 0.03],
        'loc': [
            ['1', 14705],
            ['2', 13513],
            ['3', 14492],
            ['4', 14388],
        ],
        'amp': [
            ['1', 1],
            ['2', 2],
            ['3', 5],
            ['4', 20],
        ],
        'del': [
            ['1', 400],
            ['2', 100],
            ['3', 300],
            ['4', 200],
        ]
    })
    initial = ParameterGroup.from_dict({'k': [.006, 0.003, 0.0003, 0.03]})

    e_axis = np.arange(12820, 15120, 50)
    c_axis = np.arange(0, 150, 1.5)

    sim_model = GaussianDecayModel.from_dict({
        'compartment': ["s1", "s2", "s3", "s4"],
        'dataset': {
            'dataset1': {
                'initial_concentration': [],
                'megacomplex': [],
                'kinetic': ['k.1', 'k.2', 'k.3', 'k.4'],
                'location': ['loc.1', 'loc.2', 'loc.3', 'loc.4'],
                'delta': ['del.1', 'del.2', 'del.3', 'del.4'],
                'amplitude': ['amp.1', 'amp.2', 'amp.3', 'amp.4'],
            }
        }
    })
    model = DecayModel.from_dict({
        'compartment': ["s1", "s2", "s3", "s4"],
        'dataset': {
            'dataset1': {
                'initial_concentration': [],
                'megacomplex': [],
                'kinetic': ['k.1', 'k.2', 'k.3', 'k.4']
            }
        }
    })


@pytest.mark.parametrize("index_dependend", [True, False])
@pytest.mark.parametrize("grouped", [True, False])
@pytest.mark.parametrize("suite", [
    OneCompartmentDecay,
    TwoCompartmentDecay,
    MultichannelMulticomponentDecay
])
def test_fitting(suite, index_dependend, grouped):
    model = suite.model

    def gr():
        return grouped
    model.grouped = gr

    def id():
        return index_dependend
    model.index_dependend = id

    sim_model = suite.sim_model
    est_axis = suite.e_axis
    cal_axis = suite.c_axis

    print(model.validate())
    assert model.valid()

    print(sim_model.validate())
    assert sim_model.valid()

    wanted = suite.wanted
    print(wanted)
    print(sim_model.validate(wanted))
    assert sim_model.valid(wanted)

    initial = suite.initial
    print(initial)
    print(model.validate(initial))
    assert model.valid(initial)

    dataset = simulate(sim_model, 'dataset1', wanted, {'e': est_axis, 'c': cal_axis})
    print(dataset)

    assert dataset.data.shape == (cal_axis.size, est_axis.size)

    data = {'dataset1': dataset}
    scheme = Scheme(model=model, parameter=initial, data=data, nfev=5)

    result = optimize(scheme)
    print(result.optimized_parameter)
    print(result.data['dataset1'])

    for _, param in result.optimized_parameter.all():
        assert np.allclose(param.value, wanted.get(param.full_label).value,
                           rtol=1e-1)

    resultdata = result.data["dataset1"]
    assert np.array_equal(dataset.c, resultdata.c)
    assert np.array_equal(dataset.e, resultdata.e)
    assert dataset.data.shape == resultdata.data.shape
    print(dataset.data[0, 0], resultdata.data[0, 0])
    assert np.allclose(dataset.data, resultdata.data)



"""This package contains irf items."""

from typing import List
import numpy as np

from glotaran.model import model_attribute, model_attribute_typed
from glotaran.parameter import Parameter


@model_attribute(has_type=True)
class IrfMeasured:
    """A measured IRF. The data must be supplied by the dataset."""


@model_attribute(properties={
    'center': List[Parameter],
    'width': List[Parameter],
    'scale': {'type': List[Parameter], 'allow_none': True},
    'normalize': {'type': bool, 'default': False},
    'backsweep': {'type': bool, 'default': False},
    'backsweep_period': {'type': Parameter, 'allow_none': True},
}, has_type=True)
class IrfMultiGaussian:
    """
    Represents a gaussian IRF.

    One width and one center is a single gauss.

    One center and multiple widths is a multiple gaussian.

    Multiple center and multiple widths is Double-, Triple- , etc. Gaussian.

    Parameters
    ----------

    label:
        label of the irf
    center:
        one or more center of the irf as parameter indices
    width:
        one or more widths of the gaussian as parameter index
    center_dispersion:
        polynomial coefficients for the dispersion of the
        center as list of parameter indices. None for no dispersion.
    width_dispersion:
        polynomial coefficients for the dispersion of the
        width as parameter indices. None for no dispersion.

    """
    def parameter(self, index):

        centers = self.center if isinstance(self.center, list) else [self.center]
        centers = np.asarray([c.value for c in centers])

        widths = self.width if isinstance(self.width, list) else [self.width]
        widths = np.asarray([w.value for w in widths])

        len_centers = len(centers)
        len_widths = len(widths)
        if not len_centers == len_widths:
            if not min(len_centers, len_widths) == 1:
                raise Exception(f'len(centers) ({len_centers}) not equal '
                                f'len(widths) ({len_widths}) none of is 1.')
            if len_centers == 1:
                centers = [centers[0] for _ in range(len_widths)]
                len_centers = len_widths
            else:
                widths = [widths[0] for _ in range(len_centers)]
                len_widths = len_centers

        scale = self.scale if self.scale is not None else [1.0 for _ in centers]
        scale = scale if isinstance(scale, list) else [scale]
        scale = np.asarray(scale)

        backsweep = self.backsweep

        backsweep_period = self.backsweep_period.value if self.backsweep else 0

        return centers, widths, scale, backsweep, backsweep_period

    def calculate(self, index, axis):
        center, width, scale, _, _ = self.parameter(index)
        irf = scale[0] * np.exp(-1 * (axis - center[0])**2 / (2 * width[0]**2))
        if len(center) > 1:
            for i in range(1, len(center)):
                irf += scale[i] * np.exp(-1 * (axis - center[i])**2 / (2 * width[i]**2))
        return irf


@model_attribute(properties={
    'center': Parameter,
    'width': Parameter,
}, has_type=True)
class IrfGaussian(IrfMultiGaussian):
    pass


@model_attribute_typed(types={
    'gaussian': IrfGaussian,
    'multi-gaussian': IrfMultiGaussian,
    'measured': IrfMeasured,
})
class Irf(object):
    """Represents an IRF."""



"""This package contains compartment constraint items."""

import typing
import numpy as np

from glotaran.model import model_attribute
from glotaran.parameter import Parameter, ParameterGroup

from .spectral_relations import retrieve_clps


@model_attribute(properties={
    'compartment': str,
    'interval': typing.List[typing.Tuple[float, float]],
    'target': str,
    'parameter': Parameter,
    'weight': str,
}, no_label=True)
class EqualAreaPenalty:
    """An equal area constraint adds a the differenc of the sum of a
    compartements in the e matrix in one ore more intervals to the scaled sum
    of the e matrix of one or more target compartmants to resiudal. The additional
    residual is scaled with the weight."""

    def applies(self, index: any) -> bool:
        """
        Returns true if the indexx is in one of the intervals.

        Parameters
        ----------
        index : any

        Returns
        -------
        applies : bool

        """
        def applies(interval):
            return interval[0] <= index <= interval[1]
        if isinstance(self.interval, tuple):
            return applies(self.interval)
        return any([applies(i) for i in self.interval])


def has_spectral_penalties(model: typing.Type['KineticModel']) -> bool:
    return len(model.equal_area_penalties) != 0


def apply_spectral_penalties(
        model: typing.Type['KineticModel'],
        parameter: ParameterGroup,
        clp_labels: typing.List[str],
        clps: np.ndarray,
        index: float) -> np.ndarray:

    clp_labels, clps = retrieve_clps(model, parameter, clp_labels, clps, index)

    penalties = []
    for penalty in model.equal_area_penalties:
        if penalty.applies(index):
            penalty = penalty.fill(model, parameter)
            source_idx = clp_labels.index(penalty.compartment)
            target_idx = clp_labels.index(penalty.target)
            penalties.append(
                (clps[source_idx] - penalty.parameter * clps[target_idx]) * penalty.weight
            )
    return penalties



"""A register for models"""
import typing

from glotaran.model.model import Model

_model_register = {}


def register_model(model_type: str, model: Model):
    """register_model registers a model.

    Parameters
    ----------
    model_type :
        model_type is type of the model.
    model :
        model is the model to be registered.
    """
    _model_register[model_type] = model


def known_model(model_type: str) -> bool:
    """known_model returns True if the model_type is in the register.

    Parameters
    ----------
    model_type :
        model_type is type of the model.
    """
    return model_type in _model_register


def get_model(model_type: str) -> Model:
    """get_model gets a model from the register.

    Parameters
    ----------
    model_type :
        model_type is type of the model.
    """
    return _model_register[model_type]


def known_model_names() -> typing.List[str]:
    return [name for name in _model_register]



import typing
import numpy as np
import xarray as xr


def prepare_time_trace_dataset(dataset: typing.Union[xr.DataArray, xr.Dataset],
                               weight: np.ndarray = None,
                               irf: typing.Union[np.ndarray, xr.DataArray] = None
                               ) -> xr.Dataset:
    """Prepares a time trace for global analysis.

    Parameters
    ----------
    dataset :
        The dataset.
    weight :
        A weight for the dataset.
    irf :
        An IRF for the dataset.
    """

    if isinstance(dataset, xr.DataArray):
        dataset = dataset.to_dataset(name="data")

    if'data_singular_values' not in dataset:
        l, s, r = np.linalg.svd(dataset.data)
        dataset['data_left_singular_vectors'] = \
            (('time', 'left_singular_value_index'), l)
        dataset['data_singular_values'] = (('singular_value_index'), s)
        dataset['data_right_singular_vectors'] = \
            (('right_singular_value_index', 'spectral'), r)

    if weight is not None:
        dataset['weight'] = (dataset.data.dims, weight)
        dataset['weighted_data'] = (dataset.data.dims,
                                    np.multiply(dataset.data, dataset.weight))

    if irf is not None:
        if isinstance(irf, np.ndarray):
            if not len(irf.shape) == 1:
                raise Exception("IRF with more than one dimension must be `xarray.DataArray`.")
            dataset['irf'] = (('time',), irf)
        else:
            dataset['irf'] = irf

    return dataset



import collections
import itertools
import dask.array as da
import dask.bag as db
import numpy as np

ProblemDescriptor = collections.namedtuple('ProblemDescriptor',
                                           'dataset data matrix_axis global_axis')
GroupedProblem = collections.namedtuple('GroupedProblem', 'data descriptor')
GroupedProblemDescriptor = collections.namedtuple('ProblemDescriptor', 'dataset index axis')


def create_ungrouped_bag(scheme):
    bag = {}
    for label in scheme.model.dataset:
        bag[label] = ProblemDescriptor(
            scheme.model.dataset[label],
            scheme.data[label].data,
            scheme.data[label].coords[scheme.model.matrix_dimension].values,
            scheme.data[label].coords[scheme.model.global_dimension].values,
        )
    return bag


def create_grouped_bag(scheme):
    bag = None
    datasets = None
    full_axis = None
    for label in scheme.model.dataset:
        dataset = scheme.data[label]
        global_axis = dataset.coords[scheme.model.global_dimension].values
        model_axis = dataset.coords[scheme.model.matrix_dimension].values
        if bag is None:
            bag = collections.deque(
                GroupedProblem(dataset.data.isel({scheme.model.global_dimension: i}).values,
                               [GroupedProblemDescriptor(label, value, model_axis)])
                for i, value in enumerate(global_axis)
            )
            datasets = collections.deque([label] for _, _ in enumerate(global_axis))
            full_axis = collections.deque(global_axis)
        else:
            i1, i2 = _find_overlap(full_axis, global_axis, atol=scheme.group_tolerance)

            for i, j in enumerate(i1):
                datasets[j].append(label)
                bag[j] = GroupedProblem(
                    da.concatenate(
                        [bag[j][0], dataset.data.isel({scheme.model.global_dimension: i2[i]}).values]), # noqa e501
                    bag[j][1] + [GroupedProblemDescriptor(label, global_axis[i2[i]], model_axis)]
                )

            end = i2[0] if len(i2) != 0 else 0
            for i in range(0, end):
                datasets.appendleft([label])
                full_axis.appendleft(global_axis[i])
                bag.appendleft(GroupedProblem(
                    dataset.data.isel({scheme.model.global_dimension: i}.values).values,
                    [GroupedProblemDescriptor(label, global_axis[i], model_axis)]
                ))

            start = i2[-1] + 1 if len(i2) != 0 else 0
            for i in range(start, len(global_axis)):
                datasets.append([label])
                full_axis.append(global_axis[i])
                bag.append(GroupedProblem(
                    dataset.data.isel({scheme.model.global_dimension: i}).values,
                    [GroupedProblemDescriptor(label, global_axis[i], model_axis)]
                ))

    return db.from_sequence(bag), {"".join(d): d for d in datasets if len(d) > 1}


def _find_overlap(a, b, rtol=1e-05, atol=1e-08):
    ovr_a = []
    ovr_b = []
    start_b = 0
    for i, ai in enumerate(a):
        for j, bj in itertools.islice(enumerate(b), start_b, None):
            if np.isclose(ai, bj, rtol=rtol, atol=atol, equal_nan=False):
                ovr_a.append(i)
                ovr_b.append(j)
            elif bj > ai:  # (more than tolerance)
                break  # all the rest will be farther away
            else:  # bj < ai (more than tolerance)
                start_b += 1  # ignore further tests of this item
    return (ovr_a, ovr_b)



import sys
import typing
import click

import glotaran as gta

from . import util


@click.option('--dataformat', '-dfmt', default=None,
              type=click.Choice(gta.io.reader.known_reading_formats.keys()),
              help='The input format of the data. Will be infered from extension if not set.')
@click.option('--data', '-d', multiple=True,
              type=(str, click.Path(exists=True, dir_okay=False)),
              help="Path to a dataset in the form '--data DATASET_LABEL PATH_TO_DATA'")
@click.option('--out', '-o', default=None, type=click.Path(file_okay=False),
              help='Path to an output directory.', show_default=True)
@click.option('--nfev', '-n', default=None, type=click.IntRange(min=1),
              help='Maximum number of function evaluations.', show_default=True)
@click.option('--nnls', is_flag=True,
              help='Use non-negative least squares.')
@click.option('--yes', '-y', is_flag=True,
              help="Don't ask for confirmation.")
@util.signature_analysis
def optimize_cmd(dataformat: str, data: typing.List[str], out: str, nfev: int, nnls: bool,
                 yes: bool, parameter: str, model: str, scheme: str):
    """Optimizes a model.
    e.g.:
    glotaran optimize --

    """
    if scheme is not None:
        scheme = util.load_scheme_file(scheme, verbose=True)
        if nfev is not None:
            scheme.nfev = nfev
    else:
        if model is None:
            click.echo('Error: Neither scheme nor model specified', err=True)
            sys.exit(1)
        model = util.load_model_file(model, verbose=True)

        if parameter is None:
            click.echo('Error: Neither scheme nor parameter specified', err=True)
            sys.exit(1)
        parameter = util.load_parameter_file(parameter, verbose=True)

        if len(data) == 0:
            click.echo('Error: Neither scheme nor data specified', err=True)
            sys.exit(1)
        dataset_files = {arg[0]: arg[1] for arg in data}
        datasets = {}
        for label in model.dataset:
            if label not in dataset_files:
                click.echo(f"Missing dataset for '{label}'", err=True)
                sys.exit(1)
            path = dataset_files[label]
            datasets[label] = util.load_dataset_file(path, fmt=dataformat, verbose=True)

        scheme = gta.analysis.scheme.Scheme(model=model, parameter=parameter, data=datasets,
                                            nnls=nnls, nfev=nfev)

    click.echo(scheme.validate())
    click.echo(f"Use NNLS: {scheme.nnls}")
    click.echo(f"Max Nr Function Evaluations: {scheme.nfev}")
    click.echo(f"Saving directory: is '{out if out is not None else 'None'}'")

    if yes or click.confirm('Do you want to start optimization?', abort=True, default=True):
        #  try:
        #      click.echo('Preparing optimization...', nl=False)
        #      optimizer = gta.analysis.optimizer.Optimizer(scheme)
        #      click.echo(' Success')
        #  except Exception as e:
        #      click.echo(" Error")
        #      click.echo(e, err=True)
        #      sys.exit(1)
        try:
            click.echo('Optimizing...')
            result = gta.analysis.optimize.optimize(scheme)
            click.echo('Optimization done.')
            click.echo(result.markdown(with_model=False))
            click.echo('Optimized Parameter:')
            click.echo(result.optimized_parameter.markdown())
        except Exception as e:
            click.echo(f"An error occured during optimization: \n\n{e}", err=True)
            sys.exit(1)

        if out is not None:
            try:
                click.echo(f"Saving directory is '{out}'")
                if yes or click.confirm('Do you want to save the data?', default=True):
                    paths = result.save(out)
                    click.echo(f"File saving successfull, the follwing files have been written:\n")
                    for p in paths:
                        click.echo(f"* {p}")
            except Exception as e:
                click.echo(f"An error occured during optimization: \n\n{e}", err=True)
                sys.exit(1)

        click.echo('All done, have a nice day!')



import dask
import dask.bag as db
import numpy as np


def create_index_independend_ungrouped_residual(
        scheme, parameter, problem_bag, constraint_labels_and_matrices, residual_function):

    global_dimension = scheme.model.global_dimension
    reduced_clp_labels = {}
    reduced_clps = {}
    residuals = {}
    penalties = []
    for label in problem_bag:
        data = problem_bag[label].data
        size = problem_bag[label].global_axis.size
        reduced_clp_labels[label] = constraint_labels_and_matrices[label].clp_label
        matrix = constraint_labels_and_matrices[label].matrix

        reduced_clps[label] = []
        residuals[label] = []
        for i in range(size):
            data_stripe = data.isel({global_dimension: i}).values
            clp, residual = \
                dask.delayed(residual_function, nout=2)(matrix, data_stripe)
            reduced_clps[label].append(clp)
            residuals[label].append(residual)
            penalties.append(residual)

            if callable(scheme.model.has_additional_penalty_function):
                if scheme.model.has_additional_penalty_function():
                    additional_penalty = dask.delayed(scheme.model.additional_penalty_function)(
                        parameter, reduced_clp_labels[label], reduced_clps[label], i
                    )
                    penalties.append(additional_penalty)

    penalty = dask.delayed(np.concatenate)(penalties)
    return reduced_clp_labels, reduced_clps, residuals, penalty


def create_index_dependend_ungrouped_residual(
        scheme, parameter, problem_bag, matrix_jobs, residual_function):

    global_dimension = scheme.model.global_dimension
    reduced_clp_labels = {}
    reduced_clps = {}
    residuals = {}
    penalties = []
    for label in problem_bag:
        data = problem_bag[label].data
        size = problem_bag[label].global_axis.size
        matrices = matrix_jobs[label]
        reduced_clp_labels[label] = []
        reduced_clps[label] = []
        residuals[label] = []
        for i in range(size):
            clp, residual = \
                dask.delayed(residual_function, nout=2)(
                    matrices[i][1], data.isel({global_dimension: i}).values)

            clp_label = matrices[i][0]
            reduced_clp_labels[label].append(clp_label)
            reduced_clps[label].append(clp)
            residuals[label].append(residual)
            penalties.append(residual)

            if callable(scheme.model.has_additional_penalty_function):
                if scheme.model.has_additional_penalty_function():
                    additional_penalty = dask.delayed(scheme.model.additional_penalty_function)(
                        parameter, clp_label, clp, i
                    )
                    penalties.append(additional_penalty)

    penalty = dask.delayed(np.concatenate)(penalties)
    return reduced_clp_labels, reduced_clps, residuals, penalty


def create_index_independend_grouped_residual(
        scheme, parameter, problem_bag, constraint_labels_and_matrices, residual_function):

    matrix_labels = problem_bag.pluck(1)\
        .map(lambda group: "".join(problem.dataset for problem in group))\

    def penalty_function(matrix_label, problem, labels_and_matrices):

        clp, residual = residual_function(labels_and_matrices[matrix_label].matrix, problem.data)

        penalty = residual
        if callable(scheme.model.has_additional_penalty_function):
            if scheme.model.has_additional_penalty_function():
                additional_penalty = scheme.model.additional_penalty_function(
                    parameter, labels_and_matrices[matrix_label].clp_label, clp, problem.index
                )
                penalty = np.concatenate([penalty, additional_penalty])
        return clp, residual, penalty
    penalty_bag = \
        db.map(penalty_function, matrix_labels, problem_bag, constraint_labels_and_matrices)

    reduced_clp_label = {label: constraint_labels_and_matrices[label].clp_label
                         for label in constraint_labels_and_matrices}
    reduced_clps = penalty_bag.pluck(0)
    residuals = penalty_bag.pluck(1)
    penalty = dask.delayed(np.concatenate)(penalty_bag.pluck(2))

    return reduced_clp_label, reduced_clps, residuals, penalty


def create_index_dependend_grouped_residual(
        scheme, parameter, problem_bag, constraint_labels_and_matrices, residual_function):

    def penalty_function(problem, labels_and_matrices):

        clp, residual = residual_function(labels_and_matrices.matrix, problem.data)

        penalty = residual
        if callable(scheme.model.has_additional_penalty_function):
            if scheme.model.has_additional_penalty_function():
                additional_penalty = scheme.model.additional_penalty_function(
                    parameter, labels_and_matrices.clp_label, clp, problem.index
                )
                penalty = np.concatenate([penalty, additional_penalty])
        return clp, residual, penalty
    penalty_bag = \
        db.map(penalty_function, problem_bag, constraint_labels_and_matrices)

    reduced_clp_labels = constraint_labels_and_matrices.pluck(0)
    reduced_clps = penalty_bag.pluck(0)
    residuals = penalty_bag.pluck(1)
    penalty = dask.delayed(np.concatenate)(penalty_bag.pluck(2))

    return reduced_clp_labels, reduced_clps, residuals, penalty



import numpy as np

from glotaran.model import Model, model, model_attribute


def calculate_c(dataset_descriptor=None, axis=None, index=None):
    compartments = ['s1', 's2']
    r_compartments = []
    array = np.zeros((axis.shape[0], len(compartments)))

    for i in range(len(compartments)):
        r_compartments.append(compartments[i])
        for j in range(axis.shape[0]):
            array[j, i] = (i + j) * axis[j]
    return (r_compartments, array)


def calculate_e(dataset, axis):
    compartments = ['s1', 's2']
    r_compartments = []
    array = np.zeros((axis.shape[0], len(compartments)))

    for i in range(len(compartments)):
        r_compartments.append(compartments[i])
        for j in range(axis.shape[0]):
            array[j, i] = (i + j) * axis[j]
    return (r_compartments, array)


@model_attribute(
    properties={
        'grouped': bool,
        'indexdependend': bool,
    }
)
class MockMegacomplex:
    pass


@model('mock',
       matrix=calculate_c,
       matrix_dimension='c',
       global_matrix=calculate_e,
       global_dimension='e',
       megacomplex_type=MockMegacomplex,
       )
class MockModel(Model):
    pass
